/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * \package    RobotAPI::gui-plugins::SkillManagerMonitorWidgetController
 * \author     Raphael Grimm ( raphael dot grimm at kit dot edu )
 * \date       2020
 * \copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#include "SkillManagerMonitorWidgetController.h"

#include <memory>
#include <optional>
#include <regex>
#include <string>

#include <SimoxUtility/algorithm/string.h>

#include <RobotAPI/libraries/skills/core/Skill.h>

#include "aronTreeWidget/visitors/AronTreeWidgetConverter.h"
#include "aronTreeWidget/visitors/AronTreeWidgetModalCreator.h"

// modals
#include "aronTreeWidget/modal/text/AronTreeWidgetTextInputModalController.h"

// debug
#include <QAction>
#include <QClipboard>
#include <QDoubleSpinBox>
#include <QGridLayout>
#include <QMenu>
#include <QTextBrowser>

#include <RobotAPI/libraries/aron/converter/json/NLohmannJSONConverter.h>
#include <RobotAPI/libraries/skills/core/Skill.h>
#include <RobotAPI/libraries/skills/core/SkillExecutionRequest.h>

#include "aronTreeWidget/Data.h"
#include "aronTreeWidget/modal/text/AronTreeWidgetTextInputModalController.h"
#include "aronTreeWidget/visitors/AronTreeWidgetCreator.h"
#include "aronTreeWidget/visitors/AronTreeWidgetModalCreator.h"
#include "aronTreeWidget/widgets/SkillDescriptionWidget.h"

//configSk
namespace armarx
{

    QPointer<QDialog>
    SkillManagerMonitorWidgetController::getConfigDialog(QWidget* parent)
    {
        if (!dialog)
        {
            dialog = new SimpleConfigDialog(parent);
            dialog->addProxyFinder<skills::manager::dti::SkillManagerInterfacePrx>(
                "SkillMemory", "", "SkillMem*");
        }
        return qobject_cast<SimpleConfigDialog*>(dialog);
    }

    void
    SkillManagerMonitorWidgetController::configured()
    {
        observerName = dialog->getProxyName("SkillMemory");
    }

    void
    SkillManagerMonitorWidgetController::loadSettings(QSettings* settings)
    {
        observerName = settings->value("SkillMemory", "SkillMemory").toString().toStdString();
    }

    void
    SkillManagerMonitorWidgetController::saveSettings(QSettings* settings)
    {
        settings->setValue("SkillMemory", QString::fromStdString(observerName));
    }
} // namespace armarx

// Others
namespace armarx
{
    SkillManagerMonitorWidgetController::SkillManagerMonitorWidgetController()
    {

        widget.setupUi(getWidget());

        // setup memory
        this->mem_wrapper = std::make_shared<skills::gui::SkillManagerWrapper>();

        viewer = std::make_unique<armarx::skills::gui::SkillMemoryGUI>(
            widget.treeWidgetSkillExecutions,
            widget.groupBoxSkillExecutions->layout(),

            widget.groupBoxSkills,
            widget.splitter,

            widget.groupBoxSkillDetails,
            widget.splitter,

            widget.updateWidgetLayout,

            widget.stopAllLayout,

            widget.connectionStatusLabel,

            this->mem_wrapper);

        connectSignals();
    }

    void
    SkillManagerMonitorWidgetController::connectSignals()
    {
        // disconnect
        connect(this,
                &SkillManagerMonitorWidgetController::disconnectGui,
                this->mem_wrapper.get(),
                &skills::gui::SkillManagerWrapper::disconnectMemory);
        connect(this,
                &SkillManagerMonitorWidgetController::disconnectGui,
                this->viewer.get(),
                &skills::gui::SkillMemoryGUI::disconnectGui);

        // connect
        connect(this,
                &SkillManagerMonitorWidgetController::connectGui,
                viewer.get(),
                &skills::gui::SkillMemoryGUI::connectGui);
    }

    SkillManagerMonitorWidgetController::~SkillManagerMonitorWidgetController()
    {
    }

    void
    SkillManagerMonitorWidgetController::onInitComponent()
    {
        usingProxy(observerName);
    }

    void
    SkillManagerMonitorWidgetController::onConnectComponent()
    {

        getProxy(this->memory, observerName, 1000);
        connected = true;
        mem_wrapper->connectMemory(this->memory);
        emit connectGui(observerName);
    }

    void
    SkillManagerMonitorWidgetController::onDisconnectComponent()
    {
        connected = false;
        memory = nullptr;
        emit disconnectGui();
    }

} // namespace armarx
