#include "KnownGraspProviderSegment.h"

#include <SimoxUtility/algorithm/string/string_tools.h>
#include <VirtualRobot/Grasping/GraspSet.h>
#include <VirtualRobot/XML/ObjectIO.h>

#include <ArmarXCore/core/exceptions/LocalException.h>
#include <ArmarXCore/core/logging/Logging.h>
#include <ArmarXCore/core/rapidxml/wrapper/RapidXmlReader.h>

#include <RobotAPI/libraries/armem/server/MemoryToIceAdapter.h>

namespace armarx::armem::grasping::segment
{

    KnownGraspProviderSegment::KnownGraspProviderSegment(
        armem::server::MemoryToIceAdapter& iceMemory) :
        Base(iceMemory, PROVIDER_SEGMENT_NAME, CORE_SEGMENT_NAME)
    {
    }

    void
    KnownGraspProviderSegment::init()
    {
        Base::init();

        loadMemory();
    }

    std::optional<arondto::KnownGraspInfo>
    KnownGraspProviderSegment::knownGraspInfoFromObjectInfo(const ObjectInfo& info)
    {
        std::string objectClassName = info.className();
        auto fileLocInfo = info.file(".xml", "_Grasps");
        std::filesystem::path graspFilePath = fileLocInfo.absolutePath;

        if (std::filesystem::is_regular_file(graspFilePath))
        {
            ARMARX_INFO << "loading " << graspFilePath;
            try
            {
                auto manipulationObject =
                    VirtualRobot::ObjectIO::loadManipulationObject(graspFilePath);

                if (manipulationObject == nullptr)
                {
                    ARMARX_WARNING << "Invalid file content: " << graspFilePath;
                    return std::nullopt;
                }

                arondto::KnownGraspInfo ret;
                ret.correspondingObject.memoryName = "Object";
                ret.correspondingObject.coreSegmentName = "Class";
                ret.correspondingObject.providerSegmentName = "PriorKnowledgeData";
                ret.correspondingObject.entityName = info.idStr();
                ret.xml.package = fileLocInfo.package;
                ret.xml.path = fileLocInfo.relativePath;

                for (const VirtualRobot::GraspSetPtr& graspSet :
                     manipulationObject->getAllGraspSets())
                {
                    ARMARX_CHECK_NOT_NULL(graspSet);

                    arondto::KnownGraspSet retGraspSet;

                    retGraspSet.name = graspSet->getName();

                    retGraspSet.robot = simox::alg::split(graspSet->getRobotType(), " ").front();

                    retGraspSet.endeffector = graspSet->getEndEffector();

                    VirtualRobot::GraspSetPtr preGraspSet = [&graspSet]()
                    {
                        std::vector<VirtualRobot::GraspPtr> preGrasps;
                        for (const auto& grasp : graspSet->getGrasps())
                        {
                            // check if grasp is a prepose by checking if grasp name ends with "_Prepose"
                            if (simox::alg::ends_with(grasp->getName(), PREPOSE_SUFFIX))
                            {
                                preGrasps.emplace_back(grasp);
                            }

                            // everything else is assumed to be a grasp
                        }

                        // create a new grasp set with the preposes
                        VirtualRobot::GraspSetPtr preGraspSet = graspSet->clone();
                        preGraspSet->removeAllGrasps(); // just keep the EEF info etc
                        for (const auto& preGrasp : preGrasps)
                        {
                            preGraspSet->addGrasp(preGrasp);
                        }

                        return preGraspSet;
                    }();


                    ARMARX_DEBUG << VAROUT(preGraspSet->getSize());

                    // remove all preposes from the grasp set
                    for (const auto& preGrasp : preGraspSet->getGrasps())
                    {
                        graspSet->removeGrasp(preGrasp);
                    }


                    for (const VirtualRobot::GraspPtr& grasp : graspSet->getGrasps())
                    {
                        ARMARX_CHECK_NOT_NULL(grasp);

                        arondto::KnownGrasp retGrasp;

                        retGrasp.name = grasp->getName();
                        retGrasp.quality = grasp->getQuality();
                        retGrasp.creator = grasp->getCreationMethod();
                        retGrasp.pose = grasp->getTransformation();
                        retGrasp.prepose.reset();

                        // check if grasp has a prepose by checking if grasp name ends with "_Prepose"
                        {
                            const std::string prePoseName = retGrasp.name + PREPOSE_SUFFIX;
                            
                            ARMARX_DEBUG << "Checking for prepose '" << prePoseName << "' ...";
                            
                            if (preGraspSet->hasGrasp(prePoseName))
                            {
                                retGrasp.prepose =
                                    preGraspSet->getGrasp(prePoseName)->getTransformation();

                                // remove the prepose from the set as it found its match
                                preGraspSet->removeGrasp(preGraspSet->getGrasp(prePoseName));


                                ARMARX_DEBUG << "Found prepose `" + prePoseName + "` for grasp '" << retGrasp.name << "' in set '"
                                           << retGraspSet.name << "' for obj '" << objectClassName
                                           << "' with pose \n"
                                           << retGrasp.prepose.value();
                            }
                        }

                        // check if grasp has a prepose for a grasp with a specific name, e.g., "XY_Grasp" (GRASP_OPTIONAL_SUFFIX)
                        if (// not retGrasp.prepose.has_value() and
                            simox::alg::ends_with(retGrasp.name, GRASP_OPTIONAL_SUFFIX))
                        {
                            const std::string prePoseName =
                                retGrasp.name.substr(
                                    0, retGrasp.name.size() - std::strlen(GRASP_OPTIONAL_SUFFIX)) +
                                PREPOSE_SUFFIX;

                            ARMARX_DEBUG << "Checking for prepose '" << prePoseName << "' ...";

                            if (preGraspSet->hasGrasp(prePoseName))
                            {
                                retGrasp.prepose =
                                    preGraspSet->getGrasp(prePoseName)->getTransformation();

                                // remove the prepose from the set as it found its match
                                preGraspSet->removeGrasp(preGraspSet->getGrasp(prePoseName));

                                ARMARX_DEBUG << "Found prepose `" + prePoseName + "` for grasp '" << retGrasp.name << "' in set '"
                                           << retGraspSet.name << "' for obj '" << objectClassName
                                           << "' with pose \n"
                                           << retGrasp.prepose.value();
                            }
                        }

                        ARMARX_DEBUG << "Found grasp '" << retGrasp.name << "' in set '"
                                       << retGraspSet.name << "' for obj '" << objectClassName
                                       << "' with pose \n"
                                       << retGrasp.pose;

                        retGraspSet.grasps.push_back(retGrasp);
                    }

                    // Now, check if there are any preposes left in the set. This should not have happened.
                    if (preGraspSet->getSize() > 0)
                    {
                        ARMARX_WARNING << "Found " << preGraspSet->getSize()
                                       << " preposes in the grasp set '" << retGraspSet.name
                                       << "' for obj '" << objectClassName
                                       << "' that do not have a corresponding grasp!";
                        for(const auto& preGrasp : preGraspSet->getGrasps())
                        {
                            ARMARX_WARNING << "Prepose '" << preGrasp->getName();
                        }
                    }

                    ARMARX_CHECK(ret.graspSets.count(retGraspSet.robot + "/" + retGraspSet.name) ==
                                 0)
                        << "The grasp set `" << retGraspSet.robot + "/" + retGraspSet.name
                        << "` was defined twice!";

                    ret.graspSets[retGraspSet.robot + "/" + retGraspSet.name] = retGraspSet;
                }
                return ret;
            }
            catch (...)
            {
                ARMARX_WARNING << graspFilePath << " is not a manipulation object!"
                               << GetHandledExceptionString();
                return std::nullopt;
            }
        }
        return std::nullopt;
    }

    void
    KnownGraspProviderSegment::loadMemory()
    {
        // load data from prior knowledge
        ObjectFinder objectFinder;
        const auto now = armem::Time::Now();

        const bool checkPaths = false;
        std::vector<ObjectInfo> infos = objectFinder.findAllObjects(checkPaths);

        const MemoryID providerID =
            segmentPtr->id().withProviderSegmentName(objectFinder.getPackageName());
        ARMARX_INFO << "Checking up to " << infos.size() << " object classes from '"
                    << objectFinder.getPackageName() << "' ...";

        Commit commit;
        for (ObjectInfo& info : infos)
        {
            info.setLogError(false);
            if (auto knownGraspCandidate = knownGraspInfoFromObjectInfo(info); knownGraspCandidate)
            {
                EntityUpdate& update = commit.add();
                update.entityID = providerID.withEntityName(info.id().str());
                update.entityID.timestamp = update.arrivedTime = update.referencedTime =
                    update.sentTime = now;

                update.instancesData = {knownGraspCandidate->toAron()};

                ARMARX_DEBUG << VAROUT(knownGraspCandidate->graspSets.size());
                for (const auto& gs : knownGraspCandidate->graspSets)
                {
                    ARMARX_DEBUG << VAROUT(gs.second.grasps.size());
                    for (const auto& grasp : gs.second.grasps)
                    {
                        ARMARX_DEBUG << VAROUT(grasp.name);
                    }
                }
            }
        }
        ARMARX_INFO << "Loaded " << commit.updates.size()
                    << " grasp candidates from object classes from '"
                    << objectFinder.getPackageName() << "'.";
        auto result = iceMemory.commit(commit);
        if (!result.allSuccess())
        {
            ARMARX_WARNING << "Got errors for commit: " << result.allErrorMessages();
        }
    }
} // namespace armarx::armem::grasping::segment
