/*
* This file is part of ArmarX.
*
* ArmarX is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License version 2 as
* published by the Free Software Foundation.
*
* ArmarX is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program. If not, see <http://www.gnu.org/licenses/>.
*
* @author     Fabian Peller (fabian dot peller at kit dot edu)
* @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
*             GNU General Public License
*/

// STD/STL
#include <cmath>
#include <cstdint>
#include <cstring>
#include <memory>
#include <numeric>

#include <SimoxUtility/algorithm/get_map_keys_values.h>

#include "ArmarXCore/core/logging/Logging.h"

// Header
#include "NlohmannJSONReaderWithoutTypeCheck.h"

// ArmarX
#include <RobotAPI/interface/aron/Aron.h>
#include <RobotAPI/libraries/aron/core/Exception.h>
#include <RobotAPI/libraries/aron/core/data/visitor/nlohmannJSON/NlohmannJSONVisitor.h>

#include "../../json/Data.h"


namespace armarx::aron::data::reader
{
    data::Descriptor
    NlohmannJSONReaderWithoutTypeCheck::getDescriptor(InputType& input)
    {
        return ConstNlohmannJSONVisitor::GetDescriptor(input);
    }

    void
    NlohmannJSONReaderWithoutTypeCheck::readList(const nlohmann::json& input,
                                                 std::vector<nlohmann::json>& elements,
                                                 Path& p)
    {
        elements = input.get<std::vector<nlohmann::json>>();
    }

    void
    NlohmannJSONReaderWithoutTypeCheck::readDict(const nlohmann::json& input,
                                                 std::map<std::string, nlohmann::json>& elements,
                                                 Path& p)
    {
        elements = input.get<std::map<std::string, nlohmann::json>>();
    }


    const std::map<std::string, type::matrix::ElementType> ElementTypeAsString = {
        {"unsigned char", ::armarx::aron::type::matrix::UINT8},
        {"unsigned short", ::armarx::aron::type::matrix::UINT16},
        {"unsigned int", ::armarx::aron::type::matrix::UINT32},
        {"char", ::armarx::aron::type::matrix::INT8},
        {"short", ::armarx::aron::type::matrix::INT16},
        {"int", ::armarx::aron::type::matrix::INT32},
        {"long", ::armarx::aron::type::matrix::INT64},
        {"float", ::armarx::aron::type::matrix::FLOAT32},
        {"double", ::armarx::aron::type::matrix::FLOAT64}};

    template <typename T>
    void
    readTo(const nlohmann::json& input, std::vector<unsigned char>& data)
    {
        const std::vector<T> d = input.at("data").get<std::vector<T>>();

        const std::size_t bufferLen = d.size() * sizeof(T);

        data.resize(bufferLen);
        memcpy(data.data(), d.data(), bufferLen);
    }

    void
    NlohmannJSONReaderWithoutTypeCheck::readNDArray(const nlohmann::json& input,
                                                    std::vector<int>& shape,
                                                    std::string& typeAsString,
                                                    std::vector<unsigned char>& data,
                                                    Path& p)
    {
        shape = input.at("dims").get<std::vector<int>>();

        typeAsString = input.at("type").get<std::string>();

        ARMARX_CHECK(ElementTypeAsString.count(typeAsString) > 0)
            << "Invalid element `" << typeAsString << "`. Valid elements are "
            << simox::alg::get_keys(ElementTypeAsString) << ".";

        const type::matrix::ElementType elementType = ElementTypeAsString.at(typeAsString);

        // as we need to return a vector<uchar>, we first need to read the json array as the specific type
        // and then convert it to the vector<uchar>
        switch (elementType)
        {
            case type::matrix::UINT8:
                readTo<std::uint8_t>(input, data);
                break;
            case type::matrix::UINT16:
                readTo<std::uint16_t>(input, data);
                break;
            case type::matrix::UINT32:
                readTo<std::uint32_t>(input, data);
                break;
            case type::matrix::INT8:
                readTo<std::int8_t>(input, data);
                break;
            case type::matrix::INT16:
                readTo<std::int16_t>(input, data);
                break;
            case type::matrix::INT32:
                readTo<std::int32_t>(input, data);
                break;
            case type::matrix::INT64:
                readTo<std::int64_t>(input, data);
                break;
            case type::matrix::FLOAT32:
                readTo<std::float_t>(input, data);
                break;
            case type::matrix::FLOAT64:
                readTo<std::double_t>(input, data);
                break;
        }

        ARMARX_INFO << VAROUT(typeAsString);
    }

    void
    NlohmannJSONReaderWithoutTypeCheck::readInt(const nlohmann::json& input, int& i, Path& p)
    {
        i = input;
    }

    void
    NlohmannJSONReaderWithoutTypeCheck::readLong(const nlohmann::json& input, long& i, Path& p)
    {
        i = input;
    }

    void
    NlohmannJSONReaderWithoutTypeCheck::readFloat(const nlohmann::json& input, float& i, Path& p)
    {
        i = input;
    }

    void
    NlohmannJSONReaderWithoutTypeCheck::readDouble(const nlohmann::json& input, double& i, Path& p)
    {
        i = input;
    }

    void
    NlohmannJSONReaderWithoutTypeCheck::readString(const nlohmann::json& input,
                                                   std::string& i,
                                                   Path& p)
    {
        i = input;
    }

    void
    NlohmannJSONReaderWithoutTypeCheck::readBool(const nlohmann::json& input, bool& i, Path& p)
    {
        i = input;
    }

} // namespace armarx::aron::data::reader
